@@ -20,16 +20,16 @@
 #include "qgis_core.h"
 #include "qgis_sip.h"
 #include <QMap>
-#include <Qt>
-#include <QtCore>
 #include <QFont>
 #include <QColor>
 #include <QPainter>
-#include "qgssymbol.h"
 #include "qgis.h"
 #include "qgsmapunitscale.h"
 #include "qgscolorramp.h"
 #include "qgsarrowsymbollayer.h"
+#include "qgssymbol.h"
+
+#include <QFile>
 
 class QgsExpression;
 class QgsPathResolver;
@@ -46,6 +46,9 @@ class QIcon;
 class QPixmap;
 class QPointF;
 class QSize;
+class QMimeData;
+class QgsFeatureRenderer;
+class QgsSymbolLayerId;
 
 /**
  * \ingroup core
@@ -55,14 +58,6 @@ class CORE_EXPORT QgsSymbolLayerUtils
 {
   public:
 
-    //! Editing vertex markers
-    enum VertexMarkerType
-    {
-      SemiTransparentCircle,
-      Cross,
-      NoMarker
-    };
-
     static QString encodeColor( const QColor &color );
     static QColor decodeColor( const QString &str );
 
@@ -96,6 +91,27 @@ class CORE_EXPORT QgsSymbolLayerUtils
     static QString encodeSldBrushStyle( Qt::BrushStyle style );
     static Qt::BrushStyle decodeSldBrushStyle( const QString &str );
 
+    /**
+     * Decodes a \a string representing a symbol coordinate reference mode.
+     *
+     * \param string string to decode
+     * \param ok will be set to TRUE if \a string was successfully decoded
+     * \returns decoded marker clip mode
+     *
+     * \see encodeCoordinateReference()
+     * \since QGIS 3.24
+     */
+    static Qgis::SymbolCoordinateReference decodeCoordinateReference( const QString &string, bool *ok SIP_OUT = nullptr );
+
+    /**
+     * Encodes a symbol coordinate reference mode to a string.
+     *
+     * \param coordinateReference coordinate reference mode
+     * \see decodeCoordinateReference()
+     * \since QGIS 3.24
+     */
+    static QString encodeCoordinateReference( Qgis::SymbolCoordinateReference coordinateReference );
+
     /**
      * Decodes a \a value representing an arrow head type.
      * \since QGIS 3.2
@@ -108,6 +124,46 @@ class CORE_EXPORT QgsSymbolLayerUtils
      */
     static QgsArrowSymbolLayer::ArrowType decodeArrowType( const QVariant &value, bool *ok SIP_OUT = nullptr );
 
+    /**
+     * Decodes a \a string representing a marker clip mode.
+     *
+     * \param string string to decode
+     * \param ok will be set to TRUE if \a string was successfully decoded
+     * \returns decoded marker clip mode
+     *
+     * \see encodeMarkerClipMode()
+     * \since QGIS 3.24
+     */
+    static Qgis::MarkerClipMode decodeMarkerClipMode( const QString &string, bool *ok SIP_OUT = nullptr );
+
+    /**
+     * Encodes a marker clip \a mode to a string.
+     *
+     * \see decodeMarkerClipMode()
+     * \since QGIS 3.24
+     */
+    static QString encodeMarkerClipMode( Qgis::MarkerClipMode mode );
+
+    /**
+     * Decodes a \a string representing a line clip mode.
+     *
+     * \param string string to decode
+     * \param ok will be set to TRUE if \a string was successfully decoded
+     * \returns decoded line clip mode
+     *
+     * \see encodeLineClipMode()
+     * \since QGIS 3.24
+     */
+    static Qgis::LineClipMode decodeLineClipMode( const QString &string, bool *ok SIP_OUT = nullptr );
+
+    /**
+     * Encodes a line clip \a mode to a string.
+     *
+     * \see decodeLineClipMode()
+     * \since QGIS 3.24
+     */
+    static QString encodeLineClipMode( Qgis::LineClipMode mode );
+
     /**
      * Encodes a QPointF to a string.
      * \see decodePoint()
@@ -191,7 +247,7 @@ class CORE_EXPORT QgsSymbolLayerUtils
      * \returns matching render unit
      * \see encodeSldUom()
      */
-    static QgsUnitTypes::RenderUnit decodeSldUom( const QString &str, double *scaleFactor );
+    static QgsUnitTypes::RenderUnit decodeSldUom( const QString &str, double *scaleFactor = nullptr );
 
     /**
      * Returns the size scaled in pixels according to the uom attribute.
@@ -202,8 +258,19 @@ class CORE_EXPORT QgsSymbolLayerUtils
      */
     static double sizeInPixelsFromSldUom( const QString &uom, double size );
 
-    static QString encodeScaleMethod( QgsSymbol::ScaleMethod scaleMethod );
-    static QgsSymbol::ScaleMethod decodeScaleMethod( const QString &str );
+    /**
+     * Encodes a symbol scale method to a string.
+     *
+     * \see decodeScaleMethod()
+     */
+    static QString encodeScaleMethod( Qgis::ScaleMethod scaleMethod );
+
+    /**
+     * Decodes a symbol scale method from a string.
+     *
+     * \see encodeScaleMethod()
+     */
+    static Qgis::ScaleMethod decodeScaleMethod( const QString &str );
 
     static QPainter::CompositionMode decodeBlendMode( const QString &s );
 
@@ -212,9 +279,10 @@ class CORE_EXPORT QgsSymbolLayerUtils
      * \param symbol symbol
      * \param size target pixmap size
      * \param padding space between icon edge and symbol
+     * \param shape optional legend patch shape to use for rendering the preview icon
      * \see symbolPreviewPixmap()
      */
-    static QIcon symbolPreviewIcon( const QgsSymbol *symbol, QSize size, int padding = 0 );
+    static QIcon symbolPreviewIcon( const QgsSymbol *symbol, QSize size, int padding = 0, QgsLegendPatchShape *shape = nullptr );
 
     /**
      * Returns a pixmap preview for a color ramp.
@@ -224,13 +292,16 @@ class CORE_EXPORT QgsSymbolLayerUtils
      * \param customContext render context to use when rendering symbol
      * \param selected set to TRUE to render the symbol in a selected state
      * \param expressionContext optional custom expression context
+     * \param shape optional legend patch shape to use for rendering the preview icon
      * \note Parameter customContext added in QGIS 2.6
      * \note Parameter selected added in QGIS 3.10
      * \note Parameter expressionContext added in QGIS 3.10
+     * \note Parameter shape added in QGIS 3.14
      * \see symbolPreviewIcon()
      */
     static QPixmap symbolPreviewPixmap( const QgsSymbol *symbol, QSize size, int padding = 0, QgsRenderContext *customContext = nullptr, bool selected = false,
-                                        const QgsExpressionContext *expressionContext = nullptr );
+                                        const QgsExpressionContext *expressionContext = nullptr,
+                                        const QgsLegendPatchShape *shape = nullptr );
 
     /**
      * Draws a symbol layer preview to a QPicture
@@ -238,11 +309,12 @@ class CORE_EXPORT QgsSymbolLayerUtils
      * \param units size units
      * \param size target size of preview picture
      * \param scale map unit scale for preview
+     * \param parentSymbolType since QGIS 3.22, can be used to specify the parent symbol type so that geometry generator preview icons are correctly calculated
      * \returns QPicture containing symbol layer preview
      * \see symbolLayerPreviewIcon()
      * \since QGIS 2.9
      */
-    static QPicture symbolLayerPreviewPicture( const QgsSymbolLayer *layer, QgsUnitTypes::RenderUnit units, QSize size, const QgsMapUnitScale &scale = QgsMapUnitScale() );
+    static QPicture symbolLayerPreviewPicture( const QgsSymbolLayer *layer, QgsUnitTypes::RenderUnit units, QSize size, const QgsMapUnitScale &scale = QgsMapUnitScale(), Qgis::SymbolType parentSymbolType = Qgis::SymbolType::Hybrid );
 
     /**
      * Draws a symbol layer preview to an icon.
@@ -250,10 +322,11 @@ class CORE_EXPORT QgsSymbolLayerUtils
      * \param u size units
      * \param size target size of preview icon
      * \param scale map unit scale for preview
+     * \param parentSymbolType since QGIS 3.22, can be used to specify the parent symbol type so that geometry generator preview icons are correctly calculated
      * \returns icon containing symbol layer preview
      * \see symbolLayerPreviewPicture()
      */
-    static QIcon symbolLayerPreviewIcon( const QgsSymbolLayer *layer, QgsUnitTypes::RenderUnit u, QSize size, const QgsMapUnitScale &scale = QgsMapUnitScale() );
+    static QIcon symbolLayerPreviewIcon( const QgsSymbolLayer *layer, QgsUnitTypes::RenderUnit u, QSize size, const QgsMapUnitScale &scale = QgsMapUnitScale(), Qgis::SymbolType parentSymbolType = Qgis::SymbolType::Hybrid );
 
     /**
      * Returns an icon preview for a color ramp.
@@ -269,9 +342,15 @@ class CORE_EXPORT QgsSymbolLayerUtils
      * \param ramp color ramp
      * \param size target pixmap size
      * \param padding space between icon edge and color ramp
+     * \param direction direction to render pixmap (since QGIS 3.18)
+     * \param flipDirection set to TRUE to flip the direction of the ramp. For horizontal \a directions, ramps will be
+     * rendered left to right by default. For vertical \a directions, ramps will be rendered top to bottom by default. Setting
+     * this flag to TRUE will reverse these default directions.
+     * \param drawTransparentBackground set to FALSE to disable the checkerboard effect drawn below transparent colors in the ramp
      * \see colorRampPreviewIcon()
      */
-    static QPixmap colorRampPreviewPixmap( QgsColorRamp *ramp, QSize size, int padding = 0 );
+    static QPixmap colorRampPreviewPixmap( QgsColorRamp *ramp, QSize size, int padding = 0, Qt::Orientation direction = Qt::Horizontal,
+                                           bool flipDirection = false, bool drawTransparentBackground = true );
 
     static void drawStippledBackground( QPainter *painter, QRect rect );
 
@@ -279,7 +358,7 @@ class CORE_EXPORT QgsSymbolLayerUtils
      * Draws a vertex symbol at (painter) coordinates x, y. (Useful to assist vertex editing.)
      * \since QGIS 3.4.5
      */
-    static void drawVertexMarker( double x, double y, QPainter &p, QgsSymbolLayerUtils::VertexMarkerType type, int markerSize );
+    static void drawVertexMarker( double x, double y, QPainter &p, Qgis::VertexMarkerType type, int markerSize );
 
     //! Returns the maximum estimated bleed for the symbol
     static double estimateMaxSymbolBleed( QgsSymbol *symbol, const QgsRenderContext &context );
@@ -329,13 +408,19 @@ class CORE_EXPORT QgsSymbolLayerUtils
      */
     static QString symbolProperties( QgsSymbol *symbol );
 
-    static bool createSymbolLayerListFromSld( QDomElement &element, QgsWkbTypes::GeometryType geomType, QgsSymbolLayerList &layers );
+    /**
+     * Creates a symbol layer list from a DOM \a element.
+     */
+    static bool createSymbolLayerListFromSld( QDomElement &element, QgsWkbTypes::GeometryType geomType, QList<QgsSymbolLayer *> &layers );
 
     static QgsSymbolLayer *createFillLayerFromSld( QDomElement &element ) SIP_FACTORY;
     static QgsSymbolLayer *createLineLayerFromSld( QDomElement &element ) SIP_FACTORY;
     static QgsSymbolLayer *createMarkerLayerFromSld( QDomElement &element ) SIP_FACTORY;
 
-    static bool convertPolygonSymbolizerToPointMarker( QDomElement &element, QgsSymbolLayerList &layerList );
+    /**
+     * Converts a polygon symbolizer \a element to a list of marker symbol layers.
+     */
+    static bool convertPolygonSymbolizerToPointMarker( QDomElement &element, QList<QgsSymbolLayer *> &layerList );
     static bool hasExternalGraphic( QDomElement &element );
     static bool hasWellKnownMark( QDomElement &element );
 
@@ -398,7 +483,8 @@ class CORE_EXPORT QgsSymbolLayerUtils
 
     /**
      * Create ogr feature style string for brush
-     \param fillColr fill color*/
+     * \param fillColr fill color
+    */
     static QString ogrFeatureStyleBrush( const QColor &fillColr );
 
     static void createRotationElement( QDomDocument &doc, QDomElement &element, const QString &rotationFunc );
@@ -440,8 +526,10 @@ class CORE_EXPORT QgsSymbolLayerUtils
     static QDomElement createVendorOptionElement( QDomDocument &doc, const QString &name, const QString &value );
     static QgsStringMap getVendorOptionList( QDomElement &element );
 
-    static QgsStringMap parseProperties( QDomElement &element );
-    static void saveProperties( QgsStringMap props, QDomDocument &doc, QDomElement &element );
+    //! Parses the properties from XML and returns a map
+    static QVariantMap parseProperties( const QDomElement &element );
+    //! Saves the map of properties to XML
+    static void saveProperties( QVariantMap props, QDomDocument &doc, QDomElement &element );
 
     //! Reads a collection of symbols from XML and returns them in a map. Caller is responsible for deleting returned symbols.
     static QgsSymbolMap loadSymbols( QDomElement &element, const QgsReadWriteContext &context ) SIP_FACTORY;
@@ -532,11 +620,10 @@ class CORE_EXPORT QgsSymbolLayerUtils
      * \param data mime data to parse
      * \param hasAlpha will be set to TRUE if mime data was interpreted as a color containing
      * an explicit alpha value
-     * \returns valid color if mimedata could be interpreted as a color, otherwise an
-     * invalid color
+     * \returns valid color if mimedata could be interpreted as a color, otherwise an invalid color
      * \since QGIS 2.5
      */
-    static QColor colorFromMimeData( const QMimeData *data, bool &hasAlpha );
+    static QColor colorFromMimeData( const QMimeData *data, bool &hasAlpha SIP_OUT );
 
     /**
      * Attempts to parse mime data as a list of named colors
@@ -587,7 +674,7 @@ class CORE_EXPORT QgsSymbolLayerUtils
 
     /**
      * Attempts to parse a string as a color using a variety of common formats, including hex
-     * codes, rgb and rgba strings.
+     * codes, rgb and rgba, hsl and hsla strings.
      * \param colorStr string representing the color
      * \param containsAlpha if colorStr contains an explicit alpha value then containsAlpha will be set to TRUE
      * \param strictEval set to TRUE for stricter color parsing rules
@@ -610,6 +697,20 @@ class CORE_EXPORT QgsSymbolLayerUtils
      */
     static void premultiplyColor( QColor &rgb, int alpha );
 
+    /**
+     * Attempts to condense a \a fill and \a outline layer, by moving the outline layer to the
+     * fill symbol's stroke.
+     *
+     * This will only be done if the \a outline can be transformed into a stroke on the fill layer
+     * losslessly. If so, \a fill will be updated in place with the new stroke. Any existing stroke
+     * settings in \a fill will be replaced.
+     *
+     * Returns TRUE if the fill and outline were successfully condensed.
+     *
+     * \since QGIS 3.20
+     */
+    static bool condenseFillAndOutline( QgsFillSymbolLayer *fill, QgsLineSymbolLayer *outline );
+
     //! Sorts the passed list in requested order
     static void sortVariantList( QList<QVariant> &list, Qt::SortOrder order );
     //! Returns a point on the line from startPoint to directionPoint that is a certain distance away from the starting point
@@ -639,11 +740,50 @@ class CORE_EXPORT QgsSymbolLayerUtils
     static QPointF polygonCentroid( const QPolygonF &points );
 
     //! Calculate a point on the surface of a QPolygonF
-    static QPointF polygonPointOnSurface( const QPolygonF &points, QList<QPolygonF> *rings = nullptr );
+    static QPointF polygonPointOnSurface( const QPolygonF &points, const QVector<QPolygonF> *rings = nullptr );
 
     //! Calculate whether a point is within of a QPolygonF
     static bool pointInPolygon( const QPolygonF &points, QPointF point );
 
+    /**
+     * Returns the total length of a \a polyline.
+     *
+     * \since QGIS 3.20
+     */
+    static double polylineLength( const QPolygonF &polyline );
+
+    /**
+     * Returns the substring of a \a polyline which starts at \a startOffset from the beginning of the line
+     * and ends at \a endOffset from the start of the line.
+     *
+     * If \a startOffset is less than 0, then the start point will be calculated by subtracting that distance
+     * from the end of the line. Similarly, if \a endOffset is less than zero then the end point will be subtracted
+     * from the end of the line.
+     *
+     * May return an empty linestring if the substring is zero length.
+     *
+     * \since QGIS 3.16
+     */
+    static QPolygonF polylineSubstring( const QPolygonF &polyline, double startOffset, double endOffset );
+
+    /**
+     * Returns TRUE if the angle formed by the line \a p1 - \a p2 - \a p3 forms a "sharp" corner.
+     *
+     * Sharp corners form an angle which exceeds a 45 degree threshold.
+     *
+     * \since QGIS 3.16
+     */
+    static bool isSharpCorner( QPointF p1, QPointF p2, QPointF p3 );
+
+    /**
+     * Appends a polyline \a line to an existing \a target polyline.
+     *
+     * Any duplicate points at the start \a line which match the end point from \a target will be skipped.
+     *
+     * \since QGIS 3.16
+     */
+    static void appendPolyline( QPolygonF &target, const QPolygonF &line );
+
     /**
      * Returns a new valid expression instance for given field or expression string.
      * If the input is not a valid expression, it is assumed that it is a field name and gets properly quoted.
@@ -675,33 +815,33 @@ class CORE_EXPORT QgsSymbolLayerUtils
      *  returns the value un-modified
      * \since QGIS 3.0
      */
-    static double rescaleUom( double size, QgsUnitTypes::RenderUnit unit, const QgsStringMap &props );
+    static double rescaleUom( double size, QgsUnitTypes::RenderUnit unit, const QVariantMap &props );
 
     /**
      * Rescales the given point based on the uomScale found in the props, if any is found, otherwise
      *  returns a copy of the original point
      * \since QGIS 3.0
      */
-    static QPointF rescaleUom( QPointF point, QgsUnitTypes::RenderUnit unit, const QgsStringMap &props ) SIP_PYNAME( rescalePointUom );
+    static QPointF rescaleUom( QPointF point, QgsUnitTypes::RenderUnit unit, const QVariantMap &props ) SIP_PYNAME( rescalePointUom );
 
     /**
      * Rescales the given array based on the uomScale found in the props, if any is found, otherwise
      *  returns a copy of the original point
      * \since QGIS 3.0
      */
-    static QVector<qreal> rescaleUom( const QVector<qreal> &array, QgsUnitTypes::RenderUnit unit, const QgsStringMap &props ) SIP_PYNAME( rescaleArrayUom );
+    static QVector<qreal> rescaleUom( const QVector<qreal> &array, QgsUnitTypes::RenderUnit unit, const QVariantMap &props ) SIP_PYNAME( rescaleArrayUom );
 
     /**
      * Checks if the properties contain scaleMinDenom and scaleMaxDenom, if available, they are added into the SE Rule element
      * \since QGIS 3.0
      */
-    static void applyScaleDependency( QDomDocument &doc, QDomElement &ruleElem, QgsStringMap &props );
+    static void applyScaleDependency( QDomDocument &doc, QDomElement &ruleElem, QVariantMap &props );
 
     /**
       * Merges the local scale limits, if any, with the ones already in the map, if any
       * \since QGIS 3.0
       */
-    static void mergeScaleDependencies( double mScaleMinDenom, double mScaleMaxDenom, QgsStringMap &props );
+    static void mergeScaleDependencies( double mScaleMinDenom, double mScaleMaxDenom, QVariantMap &props );
 
     /**
      * Encodes a reference to a parametric SVG into SLD, as a succession of parametric SVG using URL parameters,
@@ -719,6 +859,55 @@ class CORE_EXPORT QgsSymbolLayerUtils
      */
     static QString getSvgParametricPath( const QString &basePath, const QColor &fillColor, const QColor &strokeColor, double strokeWidth );
 
+    /**
+     * Converts a set of symbol layer id to a set of pointers to actual symbol layers carried by the feature renderer.
+     * \since QGIS 3.12
+     */
+    static QSet<const QgsSymbolLayer *> toSymbolLayerPointers( QgsFeatureRenderer *renderer, const QSet<QgsSymbolLayerId> &symbolLayerIds );
+
+    /**
+     * \brief Creates a new symbol with size restricted to min/max size if original size is out of min/max range
+     * \param s the original symbol
+     * \param minSize the minimum size in mm
+     * \param maxSize the maximum size in mm
+     * \param context the render context
+     * \param width expected width, can be changed by the function
+     * \param height expected height, can be changed by this function
+     * \return 0 if size is within minSize/maxSize range. New symbol if size was out of min/max range. Caller takes ownership
+     */
+    static QgsSymbol *restrictedSizeSymbol( const QgsSymbol *s, double minSize, double maxSize, QgsRenderContext *context, double &width, double &height );
+
+    /**
+     * Evaluates a map of properties using the given \a context and returns a variant map with evaluated expressions from the properties.
+     * \since QGIS 3.18
+     */
+    static QgsStringMap evaluatePropertiesMap( const QMap<QString, QgsProperty> &propertiesMap, const QgsExpressionContext &context );
+
+    ///@cond PRIVATE
+#ifndef SIP_RUN
+    static QgsProperty rotateWholeSymbol( double additionalRotation, const QgsProperty &property )
+    {
+      const QString exprString = property.asExpression();
+      return QgsProperty::fromExpression( QString::number( additionalRotation ) + " + (" + exprString + ')' );
+    }
+
+    static QgsProperty scaleWholeSymbol( double scaleFactor, const QgsProperty &property )
+    {
+      const QString exprString = property.asExpression();
+      return QgsProperty::fromExpression( QString::number( scaleFactor ) + "*(" + exprString + ')' );
+    }
+
+    static QgsProperty scaleWholeSymbol( double scaleFactorX, double scaleFactorY, const QgsProperty &property )
+    {
+      const QString exprString = property.asExpression();
+      return QgsProperty::fromExpression(
+               ( !qgsDoubleNear( scaleFactorX, 0.0 ) ? "tostring(" + QString::number( scaleFactorX ) + "*(" + exprString + "))" : QStringLiteral( "'0'" ) ) +
+               "|| ',' || " +
+               ( !qgsDoubleNear( scaleFactorY, 0.0 ) ? "tostring(" + QString::number( scaleFactorY ) + "*(" + exprString + "))" : QStringLiteral( "'0'" ) ) );
+    }
+#endif
+    ///@endcond
+
 };
 
 class QPolygonF;
